"""
File that contains all the common utils functionalities used by multiple submodules in the price_engine module
"""

from typing import List

import jax
import jax.numpy as jnp


def cast_arrays(array: List[jax.Array], dtype):
    """
    Casts the array to the specified dtype

    :param array: List of arrays
    :param dtype: dtype to cast the array to
    :return: List of arrays with the specified dtype
    """
    if dtype is not None:
        return [jnp.astype(el, dtype) for el in array]

    return array


def check_shape(*args):
    """
    Checks if the shapes of the input arrays are the same
    
    :param args: List of arrays
    :return: True if the shapes are the same, False otherwise
    """
    if not args:
        return False

    shape = args[0].shape

    for arg in args[1:]:
        if arg.shape != shape:
            return False

    return True



"""
Black Scholes prices for Vanilla European options
"""
import jax
import jax.numpy as jnp

from ..common import compute_undiscounted_call_prices
from ..math import d1, cum_normal, density_normal
from ..utils import cast_arrays


def bs_price(
    spots: jax.Array,
    strikes: jax.Array,
    expires: jax.Array,
    vols: jax.Array,
    discount_rates: jax.Array = None,
    are_calls: jax.Array = None,
    dtype: jnp.dtype = None,
) -> jax.Array:
    """
    Compute the option prices for european options using the Black-Scholes model.

    :param spots: (jax.Array): Array of current asset prices.
    :param strikes: (jax.Array): Array of option strike prices.
    :param expires: (jax.Array): Array of option expiration times.
    :param vols: (jax.Array): Array of option volatility values.
    :param discount_rates: (jax.Array): Array of risk-free interest rates. Defaults to None.
    :param dividend_rates: (jax.Array): Array of dividend rates. Defaults to None.
    :param are_calls: (jax.Array): Array of booleans indicating whether options are calls (True) or puts (False).
    :param dtype: (jnp.dtype): Data type of the output. Defaults to None.
    :return: (jax.Array): Array of computed option prices.
    """
    shape = spots.shape

    [spots, strikes, expires, vols] = cast_arrays(
        [spots, strikes, expires, vols], dtype
    )

    if discount_rates is None:
        discount_rates = jnp.zeros(shape, dtype=dtype)

    discount_factors = jnp.exp(discount_rates * expires)

    calls = compute_undiscounted_call_prices(
        spots, strikes, expires, vols, discount_rates
    )

    if are_calls is None:
        return calls

    puts = calls + (strikes * discount_factors) - spots
    return jnp.where(are_calls, calls, puts)


def delta_vanilla(
    spots: jax.Array,
    strikes: jax.Array,
    expires: jax.Array,
    vols: jax.Array,
    discount_rates: jax.Array = None,
    are_calls: jax.Array = None,
    dtype: jnp.dtype = None,
) -> jax.Array:
    """
    Calculate the delta of a call/put option

    :param spots:
    :param strikes:
    :param expires:
    :param vols:
    :param discount_rates:
    :param are_calls:
    :param dtype:
    :return:
    """
    [spots, strikes, expires, vols] = cast_arrays(
        [spots, strikes, expires, vols], dtype
    )

    d1s = d1(spots, strikes, vols, expires, discount_rates)

    return (
        cum_normal(d1s)
        if are_calls is None
        else jnp.where(are_calls, cum_normal(d1s), cum_normal(d1s) - 1)
    )


def gamma_vanilla(
    spots: jax.Array,
    strikes: jax.Array,
    expires: jax.Array,
    vols: jax.Array,
    discount_rates: jax.Array = None,
    dtype: jnp.dtype = None,
) -> jax.Array:
    """
    Calculate the gamma of a european option

    :param spots:
    :param strikes:
    :param expires:
    :param vols:
    :param discount_rates:
    :param dtype:
    :return:
    """
    [spots, strikes, expires, vols] = cast_arrays(
        [spots, strikes, expires, vols], dtype
    )

    d1s = d1(spots, strikes, vols, expires, discount_rates)

    return density_normal(d1s) / (spots * vols * jnp.sqrt(expires))



"""Geometric Brownian Motion class implementation"""
import jax
import jax.numpy as jnp
from jax import random


class UnivGeometricBrownianMotion:
    """
    Geometric Brownian Motion

    Represent a 1-dimensional GBM

    # Example usage:
    params = {
        's0' : 10,
        'dtype' : jnp.float32,
        'mean' : 0.1,
        'sigma': 0.3
    }

    gmb_process = GeometricBrownianMotion(**params)
    paths = gmb_process.simulate_paths(maturity=1.0, n=100, n_sim=100)
    """

    def __init__(self, s0, mean, sigma, dtype):
        if dtype is None:
            raise ValueError("dtype must not be None")

        self._dtype = dtype
        self._s0 = jnp.asarray(s0, dtype=dtype)
        self._mean = jnp.asarray(mean, dtype=dtype)
        self._sigma = jnp.asarray(sigma, dtype=dtype)

    @property
    def mean(self):
        """
        :return: Returns the mean of the GBM
        """
        return self._mean

    @property
    def sigma(self):
        """
        :return: Returns the standard deviation of the GBM
        """
        return self._sigma

    @property
    def s0(self):
        """
        :return: Returns the initial value of the GBM
        """
        return self._s0

    @property
    def dtype(self):
        """
        :return: Returns the underlying dtype of the GBM
        """
        return self._dtype

    def simulate_paths(
        self, seed: int, maturity, n: int, n_sim: int
    ) -> jax.Array:
        """
        Simulate a sample of paths from the GBM

        :param maturity: time in years
        :param n: (int): number of steps
        :param n_sim: (int): number of simulations
        :return: (jax.Array): Array containing the sample paths
        """
        key = random.PRNGKey(seed)

        dt = maturity / n

        Xt = jnp.exp(
            (self._mean - self._sigma ** 2 / 2) * dt
            + self._sigma * jnp.sqrt(dt) * random.normal(key, shape=(n_sim, n - 1)).T
        )

        Xt = jnp.vstack([jnp.ones(n_sim), Xt])

        return self._s0 * Xt.cumprod(axis=0)



"""
Common math functions used in multiple submodules in the price_engine module
"""

import jax
import jax.numpy as jnp
from jax.scipy.special import erf
from jax.scipy.stats.norm import pdf

_SQRT_2 = jnp.sqrt(2.0)


@jax.jit
def d1(spots, strikes, vols, expires, discount_rates):
    """
    Calculate the d1 term in the Black-Scholes formula

    :param spots: Current spot price of the underlying
    :param strikes: Strike price of the option
    :param vols: Volatility of the underlying
    :param expires: Time to expiration of the option
    :param discount_rates: Risk-free rate
    :return: d1 term
    """
    vol_sqrt_t = vols * jnp.sqrt(expires)

    return jnp.divide(
        (jnp.log(spots / strikes) + (discount_rates + (vols**2 / 2)) * expires),
        vol_sqrt_t,
    )


@jax.jit
def cum_normal(x):
    """
    Cumulative normal distribution function

    :param x: Input value
    :return: Cumulative normal distribution value
    """
    return (erf(x / _SQRT_2) + 1) / 2


@jax.jit
def density_normal(x):
    """
    Normal distribution function
    
    :param x: Input value
    :return: Normal distribution value
    """
    return pdf(x)



"""Option prices computed with the Black-Scholes module"""
from jaxfin.price_engine.black_scholes.vanilla_options import (
    delta_vanilla,
    bs_price,
    gamma_vanilla,
)

european_price = bs_price
delta_european = delta_vanilla
gamma_european = gamma_vanilla



"""
Common functions for the price_engine submodules
"""

import jax
import jax.numpy as jnp
from .math import d1, cum_normal


def _compute_d1_d2(spots, strikes, expires, vols, discount_rates):
    """
    Compute the d1 and d2 terms in the Black-Scholes formula

    :param spots: Current spot price of the underlying
    :param strikes: Strike price of the option
    :param expires: Time to expiration of the option
    :param vols: Volatility of the underlying
    :param discount_rates: Risk-free rate
    :return: d1 and d2 terms
    """
    vol_sqrt_t = vols * jnp.sqrt(expires)

    _d1 = d1(spots, strikes, vols, expires, discount_rates)

    return [_d1, _d1 - vol_sqrt_t]


@jax.jit
def compute_undiscounted_call_prices(spots, strikes, expires, vols, discount_rates):
    """
    Compute the undiscounted call option prices

    :param spots: Current spot price of the underlying
    :param strikes: Strike price of the option
    :param expires: Time to expiration of the option
    :param vols: Volatility of the underlying
    :param discount_rates: Risk-free rate
    :return: Undiscounted call option prices
    """
    [_d1, _d2] = _compute_d1_d2(spots, strikes, expires, vols, discount_rates)

    return cum_normal(_d1) * spots - cum_normal(_d2) * strikes


@jax.jit
def compute_discounted_call_prices(spots, strikes, expires, vols, discount_rates):
    """
    Compute the discounted call option prices
    
    :param spots: Current spot price of the underlying
    :param strikes: Strike price of the option
    :param expires: Time to expiration of the option
    :param vols: Volatility of the underlying
    :param discount_rates: Risk-free rate
    :return: Discounted call option prices
    """
    [_d1, _d2] = _compute_d1_d2(spots, strikes, expires, vols, discount_rates)

    return cum_normal(_d1) * spots - cum_normal(_d2) * strikes * jnp.exp(
        (-discount_rates) * expires
    )



import jax.numpy as jnp

from jaxfin.price_engine.black_scholes import european_price

TOL = 1e-3
DTYPE = jnp.float32


def test_one_vanilla():
    dtype = jnp.float32
    spot = jnp.array([100], dtype=dtype)
    expire = jnp.array([1.0], dtype=dtype)
    vol = jnp.array([0.3], dtype=dtype)
    strike = jnp.array([110], dtype=dtype)

    price = european_price(spot, strike, expire, vol, dtype=dtype)

    assert price[0] == 8.141014


def test_vanilla_batch_calls():
    spots = jnp.array([100, 90, 80, 110, 120], dtype=DTYPE)
    expires = jnp.array([1.0, 1.0, 1.0, 1.0, 1.0], dtype=DTYPE)
    vols = jnp.array([.3, .25, .4, .2, .1], dtype=DTYPE)
    strikes = jnp.array([120, 120, 120, 120, 120], dtype=DTYPE)

    prices = european_price(spots, strikes, expires, vols, dtype=DTYPE)
    expected = jnp.array([5.440567, 1.602787, 3.140933, 5.010391, 4.7853127])

    assert jnp.isclose(prices, expected, atol=TOL).all()


def test_vanilla_batch_mixed():
    spots = jnp.array([100, 90, 80, 110, 120], dtype=DTYPE)
    expires = jnp.array([1.0, 1.0, 1.0, 1.0, 1.0], dtype=DTYPE)
    vols = jnp.array([.3, .25, .4, .2, .1], dtype=DTYPE)
    strikes = jnp.array([120, 75, 75, 120, 120], dtype=DTYPE)
    are_calls = jnp.array([True, False, False, True, True], dtype=jnp.bool_)

    prices = european_price(spots, strikes, expires, vols,
                            are_calls=are_calls, dtype=DTYPE)
    expected = jnp.array([5.440567, 2.7794075, 9.942638, 5.010391, 4.7853127])

    assert jnp.isclose(prices, expected, atol=TOL).all()


def test_vanilla_batch_mixed_disc():
    spots = jnp.array([100, 90, 80, 110, 120], dtype=DTYPE)
    expires = jnp.array([1.0, 1.0, 1.0, 1.0, 1.0], dtype=DTYPE)
    vols = jnp.array([.3, .25, .4, .2, .1], dtype=DTYPE)
    strikes = jnp.array([120, 75, 75, 120, 120], dtype=DTYPE)
    are_calls = jnp.array([True, False, False, True, True], dtype=jnp.bool_)
    discount_rates = jnp.array([0.01, 0.0125, 0.02, 0.03, 0.05], dtype=DTYPE)

    prices = european_price(spots, strikes, expires, vols,
                            are_calls=are_calls, discount_rates=discount_rates, dtype=DTYPE)
    expected = jnp.array([5.434498, 3.7151947, 11.442871, 4.9134297, 4.2234654])

    assert jnp.isclose(prices, expected, atol=TOL).all()



"""
Black '76 prices for options on forwards and futures
"""
import jax
import jax.numpy as jnp

from ..common import compute_undiscounted_call_prices
from ..utils import cast_arrays


def black_price(
    spots: jax.Array,
    strikes: jax.Array,
    expires: jax.Array,
    vols: jax.Array,
    discount_rates: jax.Array = None,
    dividend_rates: jax.Array = None,
    are_calls: jax.Array = None,
    dtype: jnp.dtype = None,
) -> jax.Array:
    """
    Compute the option prices for european options using the Black '76 model.

    :param spots: (jax.Array): Array of current asset prices.
    :param strikes: (jax.Array): Array of option strike prices.
    :param expires: (jax.Array): Array of option expiration times.
    :param vols: (jax.Array): Array of option volatility values.
    :param discount_rates: (jax.Array): Array of risk-free interest rates. Defaults to None.
    :param dividend_rates: (jax.Array): Array of dividend rates. Defaults to None.
    :param are_calls: (jax.Array): Array of booleans indicating whether options are calls (True) or puts (False).
    :param dtype: (jnp.dtype): Data type of the output. Defaults to None.
    :return: (jax.Array): Array of computed option prices.
    """
    shape = spots.shape

    [spots, strikes, expires, vols] = cast_arrays(
        [spots, strikes, expires, vols], dtype
    )

    if discount_rates is None:
        discount_rates = jnp.zeros(shape, dtype=dtype)

    if dividend_rates is None:
        dividend_rates = jnp.zeros(shape, dtype=dtype)

    discount_factors = jnp.exp((discount_rates - dividend_rates) * expires)
    forwards = discount_factors * spots

    undiscounted_calls = compute_undiscounted_call_prices(
        forwards, strikes, expires, vols, discount_rates
    )

    if are_calls is None:
        return discount_factors * undiscounted_calls

    undiscounted_forwards = forwards - strikes
    undiscouted_puts = undiscounted_calls - undiscounted_forwards
    return discount_factors * jnp.where(are_calls, undiscounted_calls, undiscouted_puts)



import jax.numpy as jnp

from jaxfin.models.gbm import UnivGeometricBrownianMotion


class TestUnivGBM:

    def test_init(self):
        s0 = 10
        mean = 0.1
        sigma = 0.3
        dtype = jnp.float32
        gbm = UnivGeometricBrownianMotion(s0, mean, sigma, dtype)

        assert gbm.mean == mean
        assert gbm.sigma == sigma
        assert gbm.dtype == dtype
        assert gbm.s0 == s0
